/******************************************************************************/
/*                                                                            */
/* Progarm name    : DxPool.c                                                 */
/* Output name     : DxPool.dll                                               */
/* Description     : This Plug-in node will create a shared memory and can    */
/*                   and will be able to read and write data in memory        */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 2.0                                                                */
/* Author: Prafull Kumar                    Date: March 2003                  */
/*                                                                            */
/******************************************************************************/

/* Includes */
/* <malloc.h> is not available on /390 */
#ifndef __MVS__
  #include <malloc.h>
#endif
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include "DxPool.h"

#include "pluginMsg.h"
#ifdef _CCI_UNIX_
#include <stdarg.h> 
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#else
#include <windows.h>
#endif

#define SEMKEY 88888             /* Key passed into semget operation  */
#define SHMKEY 99999             /* Key passed into shmget operation  */
#define NUMSEMS 3               /* Num of sems in created sem set    */
#define BIGCOUNT 10000          /* initial value of process counter  */

#ifdef _CCI_UNIX_
  int hDxMem=-1;
  void* hDxMemView=NULL;
  int hSemaphore=-1;
#else
  HANDLE hDxMem=NULL;
  HANDLE hDxMemView=NULL;
  HANDLE hSemaphore=NULL;
#endif

#ifdef _CCI_TRACE_
static FILE *fpTrace = NULL;
#endif

/*Memroy Pages each of 4096 bytes. default is 1*/
int noOfPoolPages=1;
/*Count of number of users using the view*/
int poolUsers=0;

/******************************************************************************/
/* insAttrTblEntry                                                            */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input) - pointer to the current context                    */
/*  nodeName      (input) - node name                                         */
/*  attributeName (input) - attribute name to be inserted into table          */
/*  type          (input) - type of the attribute                             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to an attribute table entry.                                      */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for attribute and link it to the current context.     */
/******************************************************************************/

ATTRIBUTE_TBL_ENTRY *insAttrTblEntry(
  CONTEXT_STRUCTURE* context,
  CciChar*         attrName,
  int              type
){
  static char* functionName = (char *)"insAttrTblEntry()";
  ATTRIBUTE_TBL_ENTRY* entry;
  int                  attrSize;

  /*Log ("-->insAttrTblEntry\n");*/
  /*ucsLog("  insAttrTblEntry attrName ",attrName);*/
  /*Log ("type=%d\n",type);*/
  entry = (ATTRIBUTE_TBL_ENTRY *)malloc(sizeof(ATTRIBUTE_TBL_ENTRY));
  if (entry) {
    entry->next = 0;
    entry->type = type;

    switch (entry->type) {
    case CNI_TYPE_INTEGER:
      attrSize = 4;
      break;
    case CNI_TYPE_CHAR:
      attrSize = MAX_STRING_SIZE;		
		/*setting this size to one. as We dont know the size of script in advance
		will allocate the required space when setAttribute will be called.*/

    case CNI_TYPE_BOOL:
      attrSize = 4;
    default:
      /* Error: Invalid datatype */
      break;
    }

    /* Acquire and clear storage for attribute value */
    entry->valptr = malloc(attrSize);
    if (entry->valptr) memset(entry->valptr, 0, attrSize);

    /* Store the attribute name */
    CciCharCpy(entry->name, attrName);

    /* Link an existing previous element to this one */
    if (context->attrListPrevious) context->attrListPrevious->next = entry;
    else if ((context->attrListHead) == 0) context->attrListHead = entry;

    /* Save the pointer to the previous element */
    context->attrListPrevious = entry;
  }
  else {
    /* Error: Unable to allocate memory */
  }

  return(entry);
}

/******************************************************************************/
/* getInputTerminalHandle                                                     */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal for handle to be returned         */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  The terminal object for a named input terminal.                           */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Traverses the input terminal list looking for the named input terminal.    */
/* When it finds the correct terminal, it returns the handle to it.           */
/*                                                                            */
/******************************************************************************/

void *getInputTerminalHandle(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
){
  static char* functionName = (char *)"getInputTerminalHandle()";
  TERMINAL_LIST_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->inputTerminalListHead;

  Log ("-->getInputTerminalHandle\n");
  /* Scan the list for a matching terminal name */
  while (p != 0) {
    if (CciCharCmp(p->name, terminalName) == 0) {
      /* Match found; return it */
      return(p->handle);
    }
    p = (TERMINAL_LIST_ENTRY *)p->next;
  }

  /* No match was found */
  return(0);
}

/******************************************************************************/
/* insInputTerminalListEntry                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal to be inserted into list          */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to an input terminal list entry.                                  */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for input terminal entry and link it to the current   */
/* context.                                                                   */
/******************************************************************************/

TERMINAL_LIST_ENTRY *insInputTerminalListEntry(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
){
  static char* functionName = (char *)"insInputTerminalListEntry()";
  TERMINAL_LIST_ENTRY* entry;
  int                  rc;
  Log ("-->insInputTerminalListEntry\n");

  entry = (TERMINAL_LIST_ENTRY *)malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry) {

    /* This entry is the current end of the list */
    entry->next = 0;

    /* Store the terminal name */
    CciCharCpy(entry->name, terminalName);

    /* Create terminal and save its handle */
    entry->handle = cniCreateInputTerminal(&rc, context->nodeObject, (CciChar*)terminalName);

    /* Link an existing previous element to this one */
    if (context->inputTerminalListPrevious) context->inputTerminalListPrevious->next = entry;
    else if ((context->inputTerminalListHead) == 0) context->inputTerminalListHead = entry;

    /* Save the pointer to the previous element */
    context->inputTerminalListPrevious = entry;
  }
  else {
    /* Error: Unable to allocate memory */
  }

  return(entry);
}

/******************************************************************************/
/* getOutputTerminalHandle                                                    */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal for handle to be returned         */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  The terminal object for a named output terminal.                          */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Traverses the output terminal list looking for the named output terminal.  */
/* When it finds the correct terminal, it returns the handle to it.           */
/*                                                                            */
/******************************************************************************/
void *getOutputTerminalHandle(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
){
  static char* functionName = (char *)"getOutputTerminalHandle()";
  TERMINAL_LIST_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->outputTerminalListHead;

  Log ("-->getOutputTerminalHandle\n");
  /* Scan the list for a matching terminal name */
  while (p != 0) {
    if (CciCharCmp(p->name, terminalName) == 0) {

      /* Match found; return it */
      return(p->handle);
    }
    p = (TERMINAL_LIST_ENTRY *)p->next;
  }

  /* No match was found */
  return(0);
}

/******************************************************************************/
/* insOutputTerminalListEntry                                                 */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal to be inserted into list          */
/*                                                                            */
/*  Pointer to an output terminal list entry.                                 */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for output terminal entry and link it to the current  */
/* context.                                                                   */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insOutputTerminalListEntry(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
){
  static char* functionName = (char *)"insOutputTerminalListEntry()";
  TERMINAL_LIST_ENTRY* entry;
  int                  rc;

  Log ("-->insOutputTerminalListEntry\n");
  entry = (TERMINAL_LIST_ENTRY *)malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry) {

    /* This entry is the current end of the list */
    entry->next = 0;

    /* Store the terminal name */
    CciCharCpy(entry->name, terminalName);

    /* Create terminal and save its handle */
    entry->handle = cniCreateOutputTerminal(&rc, context->nodeObject, (CciChar*)terminalName);

    /* Link an existing previous element to this one */
    if (context->outputTerminalListPrevious) context->outputTerminalListPrevious->next = entry;
    else if ((context->outputTerminalListHead) == 0) context->outputTerminalListHead = entry;

    /* Save the pointer to the previous element */
    context->outputTerminalListPrevious = entry;
  }
  else {
    /* Error: Unable to allocate memory */
  }

  return(entry);
}

/******************************************************************************/
/* createNodeContext                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  factoryObject (input) - address of factory object that owns the node      */
/*                          being created.                                    */
/*  nodeName      (input) - name of node being created.                       */
/*  nodeObject    (input) - address of the node object that has just been     */
/*                          created.                                          */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to the local context.                                             */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates any plug-in context for an instance of the node object. This is    */
/* invoked by the broker whenever a node instance is constructed. Nodes are   */
/* constructed whenever a messageflow containing the node is deployed by the  */
/* broker.                                                                    */
/******************************************************************************/

CciContext* createNodeContext(
  CciFactory* factoryObject,
  CciChar*    nodeName,
  CciNode*    nodeObject
){
  static char* functionName = (char *)"createNodeContext()";
  CONTEXT_STRUCTURE* p;
  CciChar          buffer[256];
  int isMemoryAlreadyExist;
  int rc;
  CNI_BROKER_INFO_ST broker_info_st = {CNI_BROKER_INFO_ST_DEFAULT};

  Log ("-->createNodeContext\n");

  /* Allocate a pointer to the local context */
  p = (CONTEXT_STRUCTURE *)malloc(sizeof(CONTEXT_STRUCTURE));

  if (p) {

    /* Clear the context area */
    memset(p, 0, sizeof(CONTEXT_STRUCTURE));

    /* Save our node object pointer in our context */
    p->nodeObject = nodeObject;

    /* Save our node name */
    CciCharNCpy((CciChar*)&p->nodeName, nodeName, MAX_NODE_NAME_LEN);

    /* Create attributes and set default values */
    {
      const CciChar* ucsAttr = CciString("DxPoolExpression", BIP_DEF_COMP_CCSID) ;
      insAttrTblEntry(p, (CciChar*)ucsAttr, CNI_TYPE_CHAR);
      free((void *)ucsAttr) ;
    }
    {
      const CciChar* ucsAttr = CciString("messageStoreEltNotFound", BIP_DEF_COMP_CCSID) ;
      insAttrTblEntry(p, (CciChar*)ucsAttr, CNI_TYPE_BOOL);
      free((void *)ucsAttr) ;
    }
    {
      const CciChar* ucsAttr = CciString("messageInsertKeyNotFound", BIP_DEF_COMP_CCSID) ;
      insAttrTblEntry(p, (CciChar*)ucsAttr, CNI_TYPE_BOOL);
      free((void *)ucsAttr) ;
    }
    {
      const CciChar* ucsAttr = CciString("messgaeRestore", BIP_DEF_COMP_CCSID) ;
      insAttrTblEntry(p, (CciChar*)ucsAttr, CNI_TYPE_BOOL);
      free((void *)ucsAttr) ;
    }

    /* Create terminals */
    {
      const CciChar* ucsIn = CciString("in", BIP_DEF_COMP_CCSID) ;
      insInputTerminalListEntry(p, (CciChar*)ucsIn);
      free((void *)ucsIn) ;
    }
    {
      insOutputTerminalListEntry(p, (CciChar*)constOut);
    }
    {
      insOutputTerminalListEntry(p, (CciChar*)constFailure);
    }
    isMemoryAlreadyExist=initHandleOfDxPool();
    initHandleOfDxPoolView(isMemoryAlreadyExist);
    #ifdef _CCI_UNIX_
      if(hDxMem == -1 || hDxMemView == NULL)
    #else
      if(hDxMem == NULL || hDxMemView == NULL)
    #endif
      {
	  /* Error handling if shared Memory handle is NULL*/
	  Log("  hDxMem or hDxMemView is NULL\n");
	  cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "createNodeContext", constDxCatalog, DXP_CANT_INIT_MEMORY, "MQSeriesIntegrator2",0);
      cciThrowException(&rc, CCI_CONFIGURATION_EXCEPTION, __FILE__, __LINE__, "createNodeContext", constDxCatalog, DXP_CANT_INIT_MEMORY, "MQSeriesIntegrator2", (char *)0);
      }
    }
  else {
    /* Error: Unable to allocate memory */
  }

  /* Return the pointer to the local context */
#ifdef _CCI_UNIX_
	{
        struct sembuf operations[1] = {
        2,-1,SEM_UNDO       /*decerement the Process counter in sem*/
        };
        rc = semop( hSemaphore, operations, 1 );
        if (rc == -1)
          {
            Log("  >>createNodeContext: semop() failed\n");
          }
	}
#endif
  ++poolUsers;
  Log("  poolUsers %d\n",poolUsers);

  /* Setup the BROKER_INFO_ST to query our environment */
  /*broker_info_st.brokerName.buffer               = (CciChar *)malloc(MAX_STRING_SIZE);
  broker_info_st.brokerName.bufferLength         = MAX_STRING_SIZE;
  broker_info_st.executionGroupName.buffer       = (CciChar *)malloc(MAX_STRING_SIZE);
  broker_info_st.executionGroupName.bufferLength = MAX_STRING_SIZE;
  broker_info_st.messageFlowName.buffer          = (CciChar *)malloc(MAX_STRING_SIZE);
  broker_info_st.messageFlowName.bufferLength    = MAX_STRING_SIZE;
  broker_info_st.queueManagerName.buffer         = (CciChar *)malloc(MAX_STRING_SIZE);
  broker_info_st.queueManagerName.bufferLength   = MAX_STRING_SIZE;

  cniGetBrokerInfo(0, ((CONTEXT_STRUCTURE *)p)->nodeObject, &broker_info_st);
  ucsLog("  createNodeContext cniGetBrokerInfo \n",broker_info_st.messageFlowName.buffer);

  if (broker_info_st.brokerName.buffer)         free(broker_info_st.brokerName.buffer);
  if (broker_info_st.executionGroupName.buffer) free(broker_info_st.executionGroupName.buffer);
  if (broker_info_st.messageFlowName.buffer)    free(broker_info_st.messageFlowName.buffer);
  if (broker_info_st.queueManagerName.buffer)   free(broker_info_st.queueManagerName.buffer);*/

  return((CciContext*)p);
}


void deleteNodeContext(CciContext* context)
{
	DX_POOL_EXPRESSIONS* p,*tp;
	int iMessageResotre;
	int semPoolUsers;
	unsigned short *sem_array;
	int rc;

        Log("-->deleteNodeContext\n");
	--poolUsers;
	#ifdef _CCI_UNIX_
	{
	static struct sembuf operations[1] = {
	2,1,SEM_UNDO       /*increment the Process counter in sem*/
	};
        rc = semop( hSemaphore, operations, 1 );
        if (rc == -1)
          {
            Log("  >deleteNodeContext: semop() failed\n");
          }
	}
	#endif

	if(poolUsers==0)
	{
	#ifdef _CCI_UNIX_
	       semPoolUsers=semctl(hSemaphore,2,GETVAL,0);
	       Log("  semPoolUsers %d\n",BIGCOUNT+1-semPoolUsers);
	       if (semPoolUsers<0)
	       {
	       Log("  >deleteNodeContext: semctl() GETVAL failed\n");
	       }
	       if(semPoolUsers==BIGCOUNT)
	       {
	               closeHandleOfDxPoolView();
		       closeHandleOfDxPool();
		       rc = semctl( hSemaphore, 1, IPC_RMID, sem_array);
		       if (rc==-1)
		       {
		       	Log("  >deleteNodeContext: semctl() remove id failed\n");
		       }
		       Log("  >> Semaphore Deleted <<\n");
	       }
	       hSemaphore=-1;
	#else
                iMessageResotre=getBoolAttribute(context,L"messgaeRestore");
		if(iMessageResotre==1)
			storeDxPoolToFile();
		closeHandleOfDxPoolView();
		closeHandleOfDxPool();
                CloseHandle(hSemaphore);
		hSemaphore=NULL;
	#endif
	}
	p=((CONTEXT_STRUCTURE*)context)->pDxPoolExpressions;
	while(p) {
	  free(p->valuePtr);
	  tp=p->next;
	  free(p);
	  p=tp;
	}
	Log("  finish.\n");
  return;
}

/******************************************************************************/
/* getAttributeName                                                           */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context    (input)  - pointer to the current context                      */
/*  index      (input)  - index of the attribute for which the name is to be  */
/*                        retrieved                                           */
/*  buffer     (output) - address of buffer into which attribute name is      */
/*                        copied                                              */
/*  bufferSize (input)  - length of buffer                                    */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Returns the name of an attribute by a given index. It is invoked by the    */
/* broker when it requires the names of attributes supported by a particular  */
/* instance of a node. The function must gurantee to return the attribute     */
/* in a known, defined order and to return the attribute name represented by  */
/* the index parameter.                                                       */
/*                                                                            */
/******************************************************************************/
int getAttributeName(
  CciContext* context,
  int         index,
  CciChar*    buffer,
  int         bufsize
){
  static char* functionName = (char *)"getAttributeName()";
  int                  count = 0;
  ATTRIBUTE_TBL_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attrListHead;
  Log ("-->getAttributeName\n");

  while (p != 0) {
    if (count == index) {
      CciCharCpy(buffer, p->name);

      /* Normal return; attribute name returned */
      return(0);
    }
    count++;
    p = (ATTRIBUTE_TBL_ENTRY *)p->next;
  }

  /* Attribute not owned by the plugin node; defer to base */
  return(1);
}

/******************************************************************************/
/* setAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context        (input)  - pointer to the current context                  */
/*  attributeName  (input)  - attribute name for which value is to be set     */
/*  attributeValue (input) -  value to be set for the attribute               */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Sets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a configuration request is received that attempts to    */
/* set the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/******************************************************************************/
int setAttribute(
  CciContext* context,
  CciChar*    attrName,
  CciChar*    attrValue
){
  static char* functionName = (char *)"setAttribute()";
  ATTRIBUTE_TBL_ENTRY* p = ((CONTEXT_STRUCTURE *)context)->attrListHead;
  CONTEXT_STRUCTURE*     nc = (CONTEXT_STRUCTURE *)context;
  int rc,exLen;
  void *data;

  /*Log ("-->setAttribute %d\n",p->type);*/
  /*ucsLog("  setAttribute attrName ",attrName);*/
  ucsLog("  setAttribute attrValue ",attrValue);/*MUST*/

  while (p != 0) {
    if (CciCharCmp(p->name, attrName) == 0) {

      switch(p->type) {
      case CNI_TYPE_INTEGER:
        {
          const char* mbAttrValue = mbString(attrValue, BIP_DEF_COMP_CCSID);
          *((int *)p->valptr) = atoi(mbAttrValue);
          free((void *)mbAttrValue);
        }
        break;
      case CNI_TYPE_BOOL:
        {
          const char* mbAttrValue = mbString(attrValue, BIP_DEF_COMP_CCSID);
	  const char* t_false=(char *)"false";
	  if(strcmp (mbAttrValue,t_false)==0)
          {
		*((int *)(p->valptr)) = 0;
          }
	  else
   	  {
		*((int *)(p->valptr)) = 1;
#ifndef _CCI_UNIX_
	 	if(wcscmp(attrName,L"messgaeRestore")==0)
			initDxPoolFromFile();
#endif
	  }

          free((void *)mbAttrValue);
          free((void *)t_false);
        }
        break;
      case CNI_TYPE_CHAR:
        {
	  /*PRA*/
	  /*memcpy(p->valptr, attrValue, MAX_STRING_SIZE);*/
	  exLen=CciCharLen(attrValue);
	  /*data=calloc(sizeof(CciChar)*exLen,sizeof(CciChar));*/
	  data=calloc(exLen,sizeof(CciChar));
	  p->valptr=data;
	  memcpy(p->valptr, attrValue, sizeof(CciChar)*exLen);

          rc=processDxPoolscript(context);
	  if(rc!=0)/*return rc;*/
	  {
		  Log("  DxPoolExpression is wrong..\n");
	      cciThrowException(&rc, CCI_CONFIGURATION_EXCEPTION, __FILE__, __LINE__, "setAttribute", constDxCatalog, 10, "MQSeriesIntegrator2", (char *)0);
	  }
        }
        break;

      default:
        /* Error: Invalid datatype */
        break;
      }

      /* Normal return; attribute value stored by plugin */
      return(0);
    }
    p = (ATTRIBUTE_TBL_ENTRY *)p->next;
  }

  {
      const CciChar* ucsL = CciString("label", BIP_DEF_COMP_CCSID) ;
      if(CciCharCmp((CciChar*)ucsL,attrName)==0)
      {
	constLabel=(char*)mbString(attrValue,BIP_DEF_COMP_CCSID);
        /*ucsLog(" setAttribute ",attrValue);*/
      }
      free((void *)ucsL) ;
  }
  /* Attribute not owned by plugin */
  return(1);
}

/******************************************************************************/
/* getAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*  attributeName (input)  - attribute name for which value is to be          */
/*                           retrieved                                        */
/*  buffer        (output) - address of buffer into which attribute value is  */
/*                           copied                                           */
/*  bufferSize    (input)  - length of buffer                                 */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Gets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a report request is received that causes a retrieval    */
/* of the value of a node attribute. The broker will have verified that the   */
/* attribute name is valid for the node.                                      */
/******************************************************************************/
int getAttribute(
  CciContext* context,
  CciChar*    attrName,
  CciChar*    buffer,
  int         bufsize
){
  static char* functionName = (char *)"getAttribute()";
  ATTRIBUTE_TBL_ENTRY* p = ((CONTEXT_STRUCTURE *)context)->attrListHead;
  CONTEXT_STRUCTURE*     nc = (CONTEXT_STRUCTURE *)context;
  char                 string[256];

  Log ("-->getAttribute\n");
  /*Log ("  bufsize=%d\n",bufsize);*/
  /*ucsLog("  getAttribute attrName ",attrName);*/
  while (p != 0) {
    if (CciCharCmp(p->name, attrName) == 0) {

      switch(p->type) {
      case CNI_TYPE_INTEGER:
        sprintf((char*)&string, "%d", *((int*)p->valptr));
        {
          const CciChar* ucsAttrValue = CciString(string, BIP_DEF_COMP_CCSID); /*TODO*/
          CciCharNCpy(buffer, ucsAttrValue, min(bufsize, 256));
          free((void *)ucsAttrValue);
        }
        break;
      case CNI_TYPE_BOOL:
        {
		if (*((int *)(p->valptr)))
		{
			const CciChar* ucsVal = CciString("true", BIP_DEF_COMP_CCSID) ;
			CciCharCpy(buffer,ucsVal);
			free((void *) ucsVal);
		}
		else
		{
			const CciChar* ucsVal = CciString("false", BIP_DEF_COMP_CCSID) ;
			CciCharCpy(buffer,ucsVal);
			free((void *) ucsVal);
		}
        }
        break;
      case CNI_TYPE_CHAR:
        {
          /*const CciChar* ucsAttrValue = CciString((char *)p->valptr, BIP_DEF_COMP_CCSID);
          CciCharNCpy(buffer, ucsAttrValue, min(MAX_STRING_SIZE, bufsize));
          free((void *)ucsAttrValue);*/
	  memcpy(buffer, p->valptr, min(MAX_STRING_SIZE, bufsize));
        }
      default:
        /* Error: Invalid datatype */
        break;
      }

      /* Normal return; attribute value returned */
      return(0);
    }
    p = (ATTRIBUTE_TBL_ENTRY *)p->next;
  }

  /* Attribute not owned by plugin */
  return(1);
}

/******************************************************************************/
/* evaluate                                                                   */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Performs node processing. This function contains the main processing logic */
/* of the message flow node. It is invoked by the broker when a message is    */
/* received on one of the input terminals of an instance of a node object.    */
/*                                                                            */
/******************************************************************************/
void evaluate(
  CciContext* context,
  CciMessage* destinationList,
  CciMessage* exceptionList,
  CciMessage* message
){
  static char* functionName = (char *)"evaluate()";
  CCI_EXCEPTION_ST exceptionStructure = {CCI_EXCEPTION_ST_DEFAULT};
  void* sendTerminalObject;
  int rc;
  int retValue = 1;
  int attributeBufSize = 0;
  wchar_t attributeBuffer;
  CciChar terminal[10];
  wchar_t* pAttributeBuffer = &attributeBuffer;
  CciMessage*      outMsg;
  CciMessage*      sendMsg;
  CciElement*      inRootElement;
  CciElement*      outRootElement;
  char *logbuffer=NULL;
  Log ("-->evaluate\n");

  inRootElement = cniRootElement(&rc, message);
  /* Create a new message for output */
  outMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message));

  /* Get the root element of the output message */
  outRootElement = cniRootElement(&rc, outMsg);

  /* Copy the contents of the input message to the output message */
  cniCopyElementTree(&rc, inRootElement, outRootElement);

  retValue=runDxPoolscript(context,message,destinationList,exceptionList,outMsg,&logbuffer);
  Log("  retValue of runDxPoolScript %d\n",retValue);

  /* propagate the message to the output terminal */
  /* get the handle to the output terminal */
  if(retValue == DXP_SUCCESS)  
  {
        CciCharCpy(terminal, constOut);
	sendMsg=outMsg;
  }
  else
  {
        CciCharCpy(terminal, constFailure);
	sendMsg=message;

	  if(logbuffer==NULL)
	  {
		cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, constDxCatalog, retValue, "MQSeriesIntegrator2",0);
	  }
	  else
	  {
		cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, constDxCatalog, retValue, "MQSeriesIntegrator2",logbuffer,0);
		  free(logbuffer);
		  logbuffer=NULL;
	  }
  }

  sendTerminalObject = getOutputTerminalHandle(context, terminal);

  /* check to see if terminal exists and is attached */
  if (sendTerminalObject)
  {
	if (cniIsTerminalAttached(&rc, sendTerminalObject))
	{
	  if (rc == CCI_SUCCESS)
	  {
		if(retValue == DXP_SUCCESS)  
			cniFinalize(&rc, outMsg, 0);
        retValue = cniPropagate(
			                    &rc
							   ,sendTerminalObject
							   ,destinationList
							   ,exceptionList
							   ,sendMsg
							   );
		Log("  cniPropogate.. retValue=%d rc=%d\n",retValue,rc);
		if (rc != CCI_SUCCESS)
		{
		  if (rc == CCI_EXCEPTION)
		  {
			cciGetLastExceptionData(&rc, &exceptionStructure);
			cciRethrowLastException(0);
		  }
          cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, constMQSI2, DXP_MESSAGE_NOT_PROPAGATED, "Message not propagated", (char *)0);
		}
		else
		{
		  /*cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, functionName, constMQSI2, BIP2623, "Message propagated", (char *)0);*/
		}
	  }
	  else
	  {
        if (rc == CCI_EXCEPTION)
		{
          cciRethrowLastException(0);
		}
		else
		{
		  /* some other error, better log it */
          cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, constDxCatalog, DXP_MESSAGE_NOT_PROPAGATED, "Message not propagated", (char *)0);
        }
	  }
	}
  }
  else
  {
	/* Terminal does not exist */
    cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, functionName, constDxCatalog, DXP_TERMINAL_NOT_ATTACHED, "terminal not attached",terminal,0);
  }

  cniDeleteMessage(0, outMsg);
  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/******************************************************************************/
void defineDxPoolNode(
  void* factoryObject
){
  static char* functionName = (char *)"defineDxPoolNode()";
  static CNI_VFT vftable = {CNI_VFT_DEFAULT};
  int            rc = 0;
  Log ("-->defineDxPoolNode\n");

  /* Setup function table with pointers to node implementation functions */
  vftable.iFpCreateNodeContext = createNodeContext;
  vftable.iFpDeleteNodeContext = deleteNodeContext;
  vftable.iFpGetAttributeName  = getAttributeName;
  vftable.iFpSetAttribute      = setAttribute;
  vftable.iFpGetAttribute      = getAttribute;
  vftable.iFpEvaluate          = evaluate;

  /* Define a node type supported by our factory. If any errors/exceptions    */
  /* occur during the execution of this utility function, then as we have not */
  /* supplied the returnCode argument, the exception will bypass the plugin   */
  /* and be directly handled by the broker.                                   */
  cniDefineNodeClass(&rc, factoryObject, (CciChar*)constDxPoolNode, &vftable);

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  initNodeConstants                                                         */
/*                                                                            */
/******************************************************************************/
void initNodeConstants(
  void
){
  Log ("-->initNodeConstants\n");
  constDxPoolNode             = CciString(CONST_DXPOOL_NODE, BIP_DEF_COMP_CCSID);
  constPluginNodeFactory      = CciString(CONST_PLUGIN_NODE_FACTORY, BIP_DEF_COMP_CCSID);
  constOut                    = CciString(CONST_OUT, BIP_DEF_COMP_CCSID);
  constFailure                = CciString(CONST_FAILURE, BIP_DEF_COMP_CCSID);
  constDxCatalog              = (CciChar*)CciString(CONST_DXCATALOG, BIP_DEF_COMP_CCSID);
  constMQSI2                  = (CciChar*)CciString(CONST_MQSI2, BIP_DEF_COMP_CCSID);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                                       */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine.                                       */
/*                                                                            */
/* The responsibilities of the plug-in are to:                                */
/*  - create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/*  - define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/*  - define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/*  - define input and output terminals supported by each node.               */
/*  - return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */
/*                                                                            */
/******************************************************************************/
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
{
  static char* functionName = (char *)"bipGetMessageflowNodeFactory()";
  CciFactory*      factoryObject;
  int              rc = 0;
  CCI_EXCEPTION_ST exception_st = {CCI_EXCEPTION_ST_DEFAULT};

  Log ("-->bipGetMessageflowNodeFactory\n");
  cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, functionName, L"MQSeriesIntegrator2",
         0, "Plugin <DxPoolFactoryNEWDx2> loading", (char *)0);
  /* Before we proceed we need to initialise all the static constants */
  /* that may be used by the plug-in.                                 */
  initNodeConstants();

  /* Create the Node Factory for this plug-in */
  factoryObject = cniCreateNodeFactory(0, (unsigned short *)constPluginNodeFactory);
  if (factoryObject == CCI_NULL_ADDR) {
    if (rc == CCI_EXCEPTION) {
      /* Get details of the exception */
      memset(&exception_st, 0, sizeof(exception_st));
      cciGetLastExceptionData(&rc, &exception_st);

      /* Any local error handling may go here */

      /* Rethrow the exception */
      cciRethrowLastException(&rc);
    }
    /* Any further local error handling can go here */
  }
  else {
    /* Define the nodes supported by this factory */
    defineDxPoolNode(factoryObject);
  }

  /* Return address of this factory object to the broker */
  return(factoryObject);
}

/******************************************************************************/
/* Log                                                                        */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to Log the data in file c:\DxPool.log                                 */
/*                                                                            */
/******************************************************************************/
void ucsLog(char *fmt,CciChar *ucsStr)
{
#ifdef _CCI_TRACE_
	const char *str = mbString(ucsStr, BIP_DEF_COMP_CCSID);
	if (fpTrace == NULL)
	{
#ifdef _CCI_UNIX_
		fpTrace = fopen("/tmp/DxPool.log", "a");
#else
		fpTrace = fopen("c:/DxPool.log", "a");
#endif
		if (fpTrace == NULL)
		{
			return;
		}
	}
	if(constLabel!=0)
		fprintf(fpTrace, "[%s]",constLabel);
	fprintf(fpTrace, fmt);
	fprintf(fpTrace, str);
	fprintf(fpTrace, "\n");
	free((void *)str);
	fflush(fpTrace);
#endif
}

void Log(char *strFormat, ...)
{
#ifdef _CCI_TRACE_
	va_list	list;
	va_start(list, strFormat);

	if (fpTrace == NULL)
	{
#ifdef _CCI_UNIX_
		fpTrace = fopen("/tmp/DxPool.log", "a");
#else
		fpTrace = fopen("c:/DxPool.log", "a");
#endif
		if (fpTrace == NULL)
		{
			return;
		}
	}

	if(constLabel!=0)
		fprintf(fpTrace, "[%s]",constLabel);
	vfprintf(fpTrace, strFormat, list);
	fflush(fpTrace);
	va_end(list);
#endif
}

/******************************************************************************/
/* processDxPoolscript                                                       */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/* !0 - failure, wrong script                                                 */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to to store each line of the script in structure DxPoolExpressions    */
/* and check that that the script should be correctness else throw exception  */
/*                                                                            */
/******************************************************************************/
int processDxPoolscript(CONTEXT_STRUCTURE* context)
{
	int charindex=0,oldcharindex=0,times=0,temp=0;
	int rc;
	char data[60];
	char*  scriptline;

	const CciChar* ucsExpression = CciString("DxPoolExpression", BIP_DEF_COMP_CCSID) ;
	CciChar* poolScript = getAttributePointer(context,ucsExpression);
	free((void *)ucsExpression);

	/*Log("-->processDxPoolscript -[%S] length-%d\n",poolScript,CciCharLen(poolScript));*/
	if(poolScript==NULL)
	Log("  poolScript is NULL\n");
	/*ucsLog("-->processDxPoolscript ",poolScript);*/
	/*****Varification of Script***********/
	if( (rc=CciCharIndexOf(poolScript,L'*',1))>=0 ||
		(rc=CciCharIndexOf(poolScript,L'/',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'\\',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'+',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'-',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'"',1)>=0) ||
		(rc=CciCharIndexOf(poolScript,L'\'',1)>=0) )
	{
		sprintf(data,"wrong keyword(Not Supported) '%c' used in Script",poolScript[rc]);
		cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "setAttribute", constDxCatalog, DXP_SCRIPT_ERROR, "MQSeriesIntegrator2",data, (char *)0);
		return 100;
	}
	/*****Varification of Script***********/

	for(;;){
		oldcharindex=charindex;
		charindex=CciCharIndexOf(poolScript,chSemiComa,++times);
		if(charindex==-1) break;
		/*Log("  charindex/old %d %d\n",charindex,oldcharindex);*/
		scriptline=malloc(charindex-oldcharindex+1);
		memset(scriptline,0,charindex-oldcharindex+1);
		if(oldcharindex!=0) temp=1;
		/* to store data in array in char format*/
		UTOASCIICharNCpy(scriptline,poolScript+oldcharindex+temp,charindex-oldcharindex-temp);
		rc=insertDxScript(context,scriptline);
		if(rc!=0) return rc;
		/*Log("  scriptline [%s] length %d\n",scriptline,strlen(scriptline));*/
	}
	return 0;
}

/******************************************************************************/
/* getAttributePointer                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*  attrName      (input)  - attribute name for which value is to be          */
/*                           retrieved                                        */
/* Description:                                                               */
/*                                                                            */
/* Gets the Pointer of the Data of an Attirbute                               */
/******************************************************************************/
void* getAttributePointer(void *context ,const CciChar *attrName)
{
  ATTRIBUTE_TBL_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attrListHead;
  Log("-->getAttributePointer\n");
  while (p != 0)
  {
    if (CciCharCmp(p->name, attrName) == 0)
	{
      /* Normal return */
      return(p->valptr);
    }
    p = p->next;
  }

  /* Unknown attribute */
  return(NULL); 
}

int insertDxScript(CONTEXT_STRUCTURE* context,char *pScript)
{
  DX_POOL_EXPRESSIONS* entry;
  DX_POOL_EXPRESSIONS* p;
  int rc;
  unsigned int i; char* c;char* set;
  /*check for any errors. remove '\n' '\t'...*/
  /*Log("-->insertDxScript Script[%s]\n",pScript);*/
  c=pScript;
  for(i=0;i<strlen(pScript);i++)
  {
	  if(*c=='\n' || *c=='\t' || *c==' ')
	  {
		  set=c;
		  for(;*set=='\n' || *set=='\t' || *set==' ' || *set==0;set++);
		  strcpy(c,set);
	  }
	  c++;
  }
  if(strncmp(pScript,"SET",3)!=0) return 1;
  strcpy(pScript,pScript+3);
  if(strncmp(pScript,"Input",strlen("Input"))==0)
  {
	/*Log("  Input.. %s\n",pScript);*/
    cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "insertDxScript", constDxCatalog, DXP_TRY_TO_CHANGE_INPUT_MSG, "MQSeriesIntegrator2",pScript, (char *)0);
	return 2;
  }
  entry = malloc(sizeof(DX_POOL_EXPRESSIONS));

  if (entry)
  {
    entry->next = 0;
	entry->valuePtr=pScript;
	p=context->pDxPoolExpressions;
	if(p==0)
	{
		context->pDxPoolExpressions=entry;
		return 0;
	}
	while(p->next!=0) p=p->next;
	p->next=entry;
	return DXP_SUCCESS;
  }
  return DXP_FAILED;
}

/******************************************************************************/
/* initHandleOfDxPool                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is to create the shared Memory segment. If the Shared Memory */
/* already exist, it gets the handle of that segment.                         */
/*                                                                            */
/******************************************************************************/
int initHandleOfDxPool()
{
	char poolSize[20];
	int rc;
	Log("-->initHandleOfDxPool hDxMem = 0x%X\n",hDxMem);
	
	#ifdef _CCI_UNIX_
	{
		union semun {
			int val;
			struct semid_ds *buf;
			ushort *array;
			} semctl_arg;
		hSemaphore = semget( SEMKEY, NUMSEMS, 0666);
		if ( hSemaphore == -1 )
		{
			Log("  initHandleOfDxPool semget() Creating semaphore\n");
			hSemaphore = semget( SEMKEY, NUMSEMS, 0666 | IPC_CREAT | IPC_EXCL );
			if ( hSemaphore == -1 )
			{
			Log("  >initHandleOfDxPool semget() failed\n");
			}

			semctl_arg.val=BIGCOUNT;
			rc = semctl( hSemaphore, 2, SETVAL, semctl_arg);
			if(rc == -1)
			{
			Log("  >initHandleOfDxPool semctl() SETVAL failed\n");
			}
		}
		if(hDxMem!=-1)
			return 1;
	}
	#else
		hSemaphore=CreateSemaphore(NULL,1,1,DX_POOL_SEMAPHORE_NAME);
		if(hSemaphore==NULL)
		{
			Log("  initHandleOfDxPool CreateSemaphore hSemaphore is Null\n");
			CloseHandle(hSemaphore);
		}
		if(hDxMem!=NULL)
			return 1;
	#endif

	if (getenv("DXPOOL_MEMORY_PAGES") != NULL) 
	{
		strcpy(poolSize, getenv("DXPOOL_MEMORY_PAGES")); 
		noOfPoolPages=atoi(poolSize);
		if(noOfPoolPages<=0 || noOfPoolPages>5000)
			noOfPoolPages=1;
		Log("  DXPOOL_MEMORY_PAGES %s %d\n",poolSize,noOfPoolPages);
	}

	#ifdef _CCI_UNIX_
	hDxMem = shmget(SHMKEY, 0, 0666 | IPC_EXCL);
	if(hDxMem!=-1)
	{
	    Log("  shmget hDxMem = %d\n",hDxMem);
		return 1;
	}
	#else
	hDxMem = OpenFileMapping(FILE_MAP_ALL_ACCESS, 0,DX_POOL_MEMORY_NAME);
	if(hDxMem!=NULL)
	{
	    Log("  OpenFileMapping hDxMem = 0x%X\n",hDxMem);
		return 1;
	}
	#endif


	sprintf(poolSize,"%d",noOfPoolPages);
    cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, "initHandleOfDxPool", constDxCatalog, DXP_MEMORY_SIZE, "MQSeriesIntegrator2",poolSize,0);

	#ifdef _CCI_UNIX_
	hDxMem = shmget(SHMKEY, DX_POOL_MEMORY_SIZE*noOfPoolPages, 0666 | IPC_CREAT | IPC_EXCL);
	if(hDxMem!=-1)
	{
	    Log("  shmget IPC_CREAT hDxMem = %d\n",hDxMem);
	}
	#else
	hDxMem = CreateFileMapping((HANDLE)0xFFFFFFFF,NULL,PAGE_READWRITE,0,DX_POOL_MEMORY_SIZE*noOfPoolPages,DX_POOL_MEMORY_NAME);
    Log("  hDxMem = 0x%X\n",hDxMem);
	if(hDxMem == NULL)
	{
		CloseHandle(hDxMem);
	}
	#endif
	return 0;
}

/******************************************************************************/
/* initHandleOfDxPoolView                                                     */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  MemoryExist       (input) - is Memory alreay exist.                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* This function is to get to get the handle of the view of Shared Memory     */
/*                                                                            */
/******************************************************************************/
void initHandleOfDxPoolView(int MemoryExist)
{
    Log("-->initHandleOfDxPoolView hDxMemView = 0x%X\n",hDxMemView);
	if(hDxMemView!=0)		/*if hDxMemView is already initialized*/
		return;
	#ifdef _CCI_UNIX_
		hDxMemView = shmat(hDxMem, NULL, 0);
		if ( hDxMemView==NULL )
		{
		Log("  >initHandleOfDxPoolView: shmat() failed\n");
		}
	#else
		hDxMemView = MapViewOfFile(hDxMem, FILE_MAP_ALL_ACCESS, 0, 0, 0); 
		Log("  hDxMemView = 0x%X\n",hDxMemView);
		if (hDxMemView == NULL)  
		{ 
			CloseHandle(hDxMemView);
		} 
	#endif

	if(!MemoryExist)
		initDxPoolMemory();
}

/******************************************************************************/
/* initDxPoolMemory                                                           */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* to initialize the shared Memory segment                                    */
/*                                                                            */
/******************************************************************************/
void initDxPoolMemory()
{
	DX_POOL_CONTROL p;
        Log("-->initDxPoolMemory\n");
	p.poolSize			= DX_POOL_MEMORY_SIZE*noOfPoolPages;
	p.DxPoolDataBlocks	= 0;
	p.DxPoolUsedSize	= sizeof(DX_POOL_CONTROL);

	#ifdef _CCI_UNIX_
	sem_WLock();
	#else
        WaitForSingleObject(hSemaphore,INFINITE);
	#endif

	memcpy(hDxMemView,&p,sizeof(DX_POOL_CONTROL));

	#ifdef _CCI_UNIX_
	sem_WUnLock();
	#else
	ReleaseSemaphore(hSemaphore,1,NULL);
	#endif
}

/******************************************************************************/
/* closeHandleOfDxPool                                                        */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to close the handle of Shared Memory handle                           */
/*                                                                            */
/******************************************************************************/
void closeHandleOfDxPool()
{
	Log("-->closeHandleOfDxPool hDxMem %X\n",hDxMem);
        #ifdef _CCI_UNIX_
        {
               struct shmid_ds shmid_struct;
               int rc;
               if(hDxMem!=-1)
               {
                       rc = shmctl(hDxMem, IPC_RMID, &shmid_struct);
                       if (rc==-1)
                       {
                       Log("  >closeHandleOfDxPool: shmctl() failed\n");
                       }
		       Log("  >> Shared Memory Segment Deleted<<\n");
               }
               hDxMem=-1;
        }
        #else
               if(hDxMem!=NULL)
                       CloseHandle(hDxMem);
               hDxMem=NULL;
        #endif
	Log("  closeHandleOfDxPool .. hDxMem=NULL\n");
}

/******************************************************************************/
/* closeHandleOfDxPoolView                                                    */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* used to close the handle of Shared Memory view                             */
/*                                                                            */
/******************************************************************************/
void closeHandleOfDxPoolView()
{
	Log("-->closeHandleOfDxPoolView hDxMemView %X\n",hDxMemView);
        #ifdef _CCI_UNIX_
        {
               int rc;
               rc = shmdt(hDxMemView);
               if (rc==-1)
               {
               Log("  >closeHandleOfDxPoolView: shmdt() failed\n");
               }
        }
        #else
               if(hDxMemView!=NULL)
                       UnmapViewOfFile(hDxMemView);
        #endif
        hDxMemView=NULL;
	Log("  closeHandleOfDxPoolView .. hDxMemView=NULL\n");
}

int getOperandsOfPoolScript(char* s,char* l,char* r)
{
	int end;
    Log("-->getOperandsOfPoolScript\n");
	end=CharIndexOf(s,'=',1);
	strncpy(l,s,end); 	*(l+end)=0;
	strncpy(r,s+end+1,strlen(s)-end-1);	*(r+strlen(s)-end-1)=0;
	Log("  l=[%s],r=[%s]\n",l,r);
	return 0;
}

/******************************************************************************/
/* runDxPoolscript                                                            */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  message         (input) - address of the input message object             */
/*  exceptionList   (input) - address of the exceptionList message object     */
/*  destinationList (input) - address of the destinationList message object   */
/*  outMsg          (in/out)- address of the output message object            */
/*  logbuff         (output)- address of the to log by CciLog                 */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/* !0 - failure, reason code of failure                                       */
/*                                                                            */
/* Description:                                                               */
/* This script is called to execute all the DxPool Expression.                */
/*                                                                            */
/******************************************************************************/
int  runDxPoolscript(CONTEXT_STRUCTURE* context,CciMessage* message,
				 CciMessage* exceptionList,CciMessage* destinationList,CciMessage* outMsg,char** logbuff)
{
  DX_POOL_DATA   pool;
  DX_POOL_EXPRESSIONS* p;
  CciElement *element;
  char lOperand[MAX_STRING_SIZE];
  char rOperand[MAX_STRING_SIZE];
  int rc,len,alen;
  /*CciChar elementValue[256];*/
  CciChar *elementValue;
  CciChar *ev;
  char* charElementValue=NULL;
  CciMessage* realDests = destinationList;
  CciMessage* realExcs = exceptionList;
  CciMessage* realMsg = message;
  CciElement* foundElt;
  char* pDataToUse=NULL;
  int iNullShmAllow,iNullEltAllow;
  Log("-->runDxPoolscript\n");

  {
  const CciChar* ucsAttr = CciString("messageStoreEltNotFound", BIP_DEF_COMP_CCSID) ;
  iNullShmAllow=getBoolAttribute(context,ucsAttr);
  Log("  iNullShmAllow=%d\n",iNullShmAllow);
  free((void *)ucsAttr) ;
  }
 
  {
  const CciChar* ucsAttr = CciString("messageInsertKeyNotFound", BIP_DEF_COMP_CCSID) ;
  iNullEltAllow=getBoolAttribute(context,ucsAttr);
  Log("  iNullEltAllow=%d\n",iNullEltAllow);
  free((void *)ucsAttr) ;
  }

  pool.data=NULL;

  p=context->pDxPoolExpressions;
  while(p) {
	  Log("  COMMAND ->[%s]\n",p->valuePtr);
	  getOperandsOfPoolScript(p->valuePtr,lOperand,rOperand);
	  if(strncmp(rOperand,"DxPool",6)==0)
	  {
		rc=getDxPoolDataBlock(&pool,rOperand);
		if(iNullEltAllow==0 && rc!=0)
		{
			*logbuff=strdup(rOperand);
			return DXP_KEY_NOT_FOUND;
		}
		if(iNullEltAllow==1 && rc!=0)
		{
			pool.data=malloc(1);
			strcpy(pool.data,"");
		}
		pDataToUse=pool.data;
	  }
	  else
	  {
		/*Log("    calling NodeGetValue\n");*/
		if(strncmp(rOperand,"Output",strlen("Output"))==0)
		{
			strcpy(rOperand,rOperand+strlen("Output"));
			Log("  Output.. %s\n",rOperand);
			element=NodeGetValue(rOperand,outMsg,realExcs,realDests);
		}
		else 
		{
			if (strncmp(rOperand,"Input",strlen("Input"))==0)
				strcpy(rOperand,rOperand+strlen("Input"));
			/*Log("  Input.. %s\n",rOperand);*/
			element=NodeGetValue(rOperand,realMsg,realExcs,realDests);
		}
		if(element)
		{
			len=256;
			elementValue=malloc(sizeof(CciChar)*(len+1));
			alen=cniElementCharacterValue(&rc,element,(CciChar*)elementValue,len);
			if(alen>len)
			{
				free(elementValue);
				elementValue=malloc(sizeof(CciChar)*(alen+1));
				cniElementCharacterValue(&rc,element,(CciChar*)elementValue,alen);
			}
			Log("  cniElementCharacterValue [%S]\n",elementValue);
			/*charElementValue=UnicodeToASCII(elementValue);UNIX*/
			{
			  const char* mbEltValue = mbString(elementValue, BIP_DEF_COMP_CCSID);
			  charElementValue=(char*)mbEltValue;
			}
			pDataToUse=charElementValue;
			free(elementValue);
		}
		else
		{
			Log("  runDxPoolscript-element not found\n");
			if(iNullShmAllow)
			{
				charElementValue=malloc(1);
				strcpy(charElementValue,"");
				pDataToUse=charElementValue;
			}
			else
			{
				*logbuff=strdup(rOperand);
				return DXP_ELEMENT_NOT_FOUND;
			}
		}
	  }
	  if(strncmp(lOperand,"DxPool",6)==0)
	  {
		/*Log("  calling storeDxPoolDataBlock\n");*/
		pool.dataKey=lOperand;
		pool.data=pDataToUse;
		rc=storeDxPoolDataBlock(&pool);

		if(rc!=0) return rc;
		/*should return 1, if no store .. memory is full..??*/
	  }
	  else
	  {
		if(strncmp(lOperand,"Output",strlen("Output"))==0)
		{
			strcpy(lOperand,lOperand+strlen("Output"));
			Log("  lOperand Output.. %s\n",lOperand);
		}
		/*Log("  calling NodeAddOrGetElement\n");*/
		foundElt=NodeAddOrGetElement(lOperand,outMsg,exceptionList,destinationList);
		if(foundElt!=NULL)
		{
			/*Log("  element is not NULL... calling cniSetElementCharacterValue\n");*/
			{
			      const CciChar* ucsev = CciString(pDataToUse, BIP_DEF_COMP_CCSID) ;
			      ev=(CciChar*)ucsev;
			}
			/*ev=ASCIIToUnicode(pDataToUse);UNIX*/
			Log("  data [%S]\n",ev);
		        cniSetElementCharacterValue(&rc,foundElt,ev,CciCharLen(ev));
			Log("  going to free the ev !!\n");
			if(ev!=NULL) {free(ev);ev=NULL;}
		}
		else
		{
			Log("  >element is NULL\n");
			free(pDataToUse);
			*logbuff=strdup(lOperand);
			return DXP_CANT_CREATE_ELEMENT;
		}
	  }/*main else*/
	  free(pDataToUse);
	  p=p->next;
  }
  return 0;
}

/******************************************************************************/
/* getDxPoolDataBlock                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  d        (input)  - pointer to DX_POOL_DATA, to retrieve the information  */
/*  inkey    (input)  - key value to get                                      */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown key                                                  */
/*                                                                            */
/* Description:                                                               */
/*  Search the Shared Memory to fill the structure and return it back         */
/******************************************************************************/
int getDxPoolDataBlock(DX_POOL_DATA* d,char* inkey)
{
	int i;
	char tKey[MAX_KEY_SIZE];
	void *p;
	char* data=NULL;
	DX_POOL_CONTROL c;
	Log("-->getDxPoolDataBlock\n");

	#ifdef _CCI_UNIX_
	sem_RLock();
	#else
        WaitForSingleObject(hSemaphore,INFINITE);
	#endif

	memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
	p=(char*)hDxMemView+sizeof(DX_POOL_CONTROL);
	Log("  getDxPoolDataBlock-%d blocks\n",c.DxPoolDataBlocks);
	for(i=0;i<c.DxPoolDataBlocks;i++)
	{
		memcpy(d,p,3*sizeof(int));
		/*Log("  getDxPoolDataBlock keylength=%d datalength=%d\n",d->keyLength,d->dataLength);*/
		/*Log("  getDxPoolDataBlock 0x%X %d\n",(char*)p+3*sizeof(int),min(d->keyLength,MAX_KEY_SIZE));*/
		memcpy(tKey,(char*)p+(3*sizeof(int)),min(d->keyLength,MAX_KEY_SIZE));
		/*Log("  tKey [%s]--nextIncrement %d\n",tKey,d->totalLength);*/
		if(strcmp(tKey,inkey)==0)
		{
			data=malloc(d->dataLength);
			memcpy(data,(char*)p+3*sizeof(int)+d->keyLength,d->dataLength);
			d->data=data;
			/*Log("  getDxPoolDataBlock [%s]\n",d->data);*/

			#ifdef _CCI_UNIX_
			sem_RUnLock();
			#else
			ReleaseSemaphore(hSemaphore,1,NULL);
			#endif

			return DXP_SUCCESS;
		}
		p=(char*)p+d->totalLength;
	}
	Log("  Key Not FOUND?\n");

	#ifdef _CCI_UNIX_
	sem_RUnLock();
	#else
	ReleaseSemaphore(hSemaphore,1,NULL);
	#endif

	return DXP_KEY_NOT_FOUND;
}

/******************************************************************************/
/* storeDxPoolDataBlock                                                       */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  d        (input)  - pointer to DX_POOL_DATA, to store the information     */
/*                                                                            */
/* Returns:                                                                   */
/*   NULL                                                                     */
/*                                                                            */
/* Description:                                                               */
/*  Stores the data to the shared Memory.                                     */
/******************************************************************************/
int storeDxPoolDataBlock(DX_POOL_DATA* d)
{
	DX_POOL_CONTROL c;
	DX_POOL_DATA data;
	int i,size=0;
	char tKey[MAX_KEY_SIZE];
	void *p;
	
	Log("-->storeDxPoolDataBlock\n");
	size=sizeof(DX_POOL_CONTROL);

	#ifdef _CCI_UNIX_
	sem_WLock();
	#else
        WaitForSingleObject(hSemaphore,INFINITE);
	#endif
	
	memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
	Log("  ####Shm Control %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);
	if(c.DxPoolUsedSize>DX_POOL_MEMORY_SIZE*noOfPoolPages)
	{
		Log("  >>Pool Memory UsedSize is more than abl memory\n");
		#ifdef _CCI_UNIX_
		sem_WUnLock();
		#else
		ReleaseSemaphore(hSemaphore,1,NULL);
		#endif
		return DXP_MEMORY_FULL;
	}

	d->keyLength=strlen(d->dataKey)+1;
	d->dataLength=strlen(d->data)+1;
	d->totalLength=3*sizeof(int)+d->keyLength+d->dataLength;

	p=(char*)hDxMemView+sizeof(DX_POOL_CONTROL);
	/*check if the key already exist.*/
	for(i=0;i<c.DxPoolDataBlocks;i++)
	{
		memcpy(&data,p,3*sizeof(int));
		/*Log("  storeDxPoolDataBlock TOTAL=%d keylength=%d datalength=%d\n",data.totalLength,data.keyLength,data.dataLength);*/
		/*Log("  storeDxPoolDataBlock 0x%X %d\n",(char*)p+3*sizeof(int),min(data.keyLength,MAX_KEY_SIZE));*/
		memcpy(tKey,(char*)p+(3*sizeof(int)),min(data.keyLength,MAX_KEY_SIZE));
		size+=data.totalLength;
		if(strcmp(tKey,d->dataKey)==0)
		{
                        /*Checking if inserting the element call full the Shared Memory*/
                        /*Will return from here. It will not swap the earlier memory*/
                        if(c.DxPoolUsedSize-data.totalLength+d->totalLength>(DX_POOL_MEMORY_SIZE*noOfPoolPages))
                        {
                                Log("  >>>> Key Found!! Pool Full\n");
                                #ifdef _CCI_UNIX_
                                sem_WUnLock();
                                #else
                                ReleaseSemaphore(hSemaphore,1,NULL);
                                #endif

                                return DXP_MEMORY_FULL;
                        }

			memcpy((char*)p,(char*)p+data.totalLength,c.DxPoolUsedSize-size);
			/*Log("  #### 0x%X 0x%X %d\n",p,(char*)p+data.totalLength,c.DxPoolUsedSize-size);*/

			c.DxPoolDataBlocks--;
			c.DxPoolUsedSize-=data.totalLength;
			/*Log("  #### %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);*/
			break;
		}
		p=(char*)p+data.totalLength;
	}
	Log("  Used Pool Memory %d out of %d. Want to insert %d bytes\n",c.DxPoolUsedSize,(DX_POOL_MEMORY_SIZE*noOfPoolPages),d->totalLength);
	if(c.DxPoolUsedSize+d->totalLength>(DX_POOL_MEMORY_SIZE*noOfPoolPages))
	{
		Log("  >>>>>>>Pool Full\n");
		/**If Its going out .. save the Control before exiting...UNIX**********/
		Log("  ####Saving control(memory full) %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);
		memcpy(hDxMemView,&c,sizeof(DX_POOL_CONTROL));
		/**********************************************/
		#ifdef _CCI_UNIX_
		sem_WUnLock();
		#else
		ReleaseSemaphore(hSemaphore,1,NULL);
		#endif

		return DXP_MEMORY_FULL;
	}
	memcpy((char*)hDxMemView+c.DxPoolUsedSize,d,sizeof(int)*3);
	memcpy((char*)hDxMemView+c.DxPoolUsedSize+sizeof(int)*3,d->dataKey,d->keyLength);
	memcpy((char*)hDxMemView+c.DxPoolUsedSize+sizeof(int)*3+d->keyLength,d->data,d->dataLength);
	c.DxPoolUsedSize+=d->totalLength;
	c.DxPoolDataBlocks++;
	Log("  %d %d %d [%s] [%s]\n",d->totalLength,d->keyLength,d->dataLength,d->dataKey,d->data);
	memcpy(hDxMemView,&c,sizeof(DX_POOL_CONTROL));

	#ifdef _CCI_UNIX_
	sem_WUnLock();
	#else
	ReleaseSemaphore(hSemaphore,1,NULL);
	#endif

	return 0;
}

/******************************************************************************/
/* getBoolAttribute                                                           */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*  attrName      (input)  - attribute name for which value is to be          */
/*                           retrieved                                        */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - bool attribute not set                                                */
/*  1 - bool attribute is set                                                 */
/* -1 - bool attribute not found                                              */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Gets the value of an boolean attribute                                     */
/******************************************************************************/
int getBoolAttribute(void *context,const CciChar *attrName)
{
  ATTRIBUTE_TBL_ENTRY *p = ((CONTEXT_STRUCTURE *)context)->attrListHead;
  int val=-1;
  Log("-->getBoolAttribute\n");
  while (p != 0)
  {
    /*ucsLog("  getBoolAttribute->",p->name);*/
    if (CciCharCmp(p->name, attrName) == 0)
	{
      switch(p->type)
	  {
        case CNI_TYPE_BOOL:
		  val=*((int *)(p->valptr));
	          /*Log("  getBoolAttribute %d\n",val);*/
          break;
        default:
          break;
      }

      /* Normal return */
      return(val);
    }
    p = p->next;
  }

  /* Unknown attribute */
  return(-1); 
}


#ifdef _CCI_UNIX_
void sem_WLock()
{
/*
	3 Semaphore in the Set
	0- Writer
	1- Reader
	2- #of Processes
*/
	int rc;
	static struct sembuf operations[3] = {
	1,0,0,       /*Wait for all the readers to finish*/
	0,0,0,	     /*Wait for any Writer to finish*/
	0,1,SEM_UNDO /*Lock the Writer*/
	};
    Log("-->sem_WLock\n");

    /* First, check to see if the second semaphore is a zero.  If it*/
    /* isn't, it is busy right now.  The semop() command will wait  */
    /* for the semaphore to reach zero before executing the semop() */
    /* When it is zero, increment the first semaphore to show that  */
    /* the shared memory segment is busy.                           */

        rc = semop( hSemaphore, operations, 3 );
        if (rc == -1)
          {
            Log("  >sem_WLock: semop() failed\n");
          }
}

void sem_WUnLock()
{
	int rc;
	static struct sembuf operations[1] = {
	0,-1,SEM_UNDO       /*release the Writer Lock*/
	};
    Log("-->sem_WUnLock\n");

    /* Release the shared memory segmnent by decrementing the in use */
    /* semaphore (the first one).  Increment the second semaphore to */
    /* show that the client is finished with it.                     */

        rc = semop( hSemaphore, operations, 1 );
        if (rc == -1)
          {
            Log("  >sem_WUnLock: semop() failed\n");
          }
}

void sem_RLock()
{
	int rc;
	static struct sembuf operations[2] = {
	0,0,0,       /*Wait for all the Writers to finish*/
	1,1,SEM_UNDO /*increment the reader*/
	};
    Log("-->sem_RLock\n");

        rc = semop( hSemaphore, operations, 2 );
        if (rc == -1)
          {
            Log("  >sem_RLock: semop() failed\n");
          }
}

void sem_RUnLock()
{
	int rc;
	static struct sembuf operations[1] = {
	1,-1,SEM_UNDO       /*release the Reader Lock*/
	};
    Log("-->sem_RUnLock\n");

        rc = semop( hSemaphore, operations, 1 );
        if (rc == -1)
          {
            Log("  >sem_RUnLock: semop() failed\n");
          }
}
#endif


/******************************************************************************/
/* storeDxPoolToFile                                                          */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None                                                                      */
/*                                                                            */
/* Returns:                                                                   */
/*  0 - success                                                               */
/*  1 - failure                                                               */
/*                                                                            */
/* Description:                                                               */
/*  Stores the data of shared Memory to a file c:\DxPoolData.bin              */
/******************************************************************************/
int storeDxPoolToFile()
{
  DX_POOL_CONTROL c;
  FILE *fp=NULL;
  int ret;
  Log("-->storeDxPoolToFile\n");
  if(hDxMemView==NULL) return DXP_FAILED;

#ifdef _CCI_UNIX_
  fp = fopen("/tmp/DxPoolData.bin","wb");
#else
  fp = fopen("c:\\DxPoolData.bin","wb");
#endif
  if(fp==NULL) return DXP_FAILED;

  #ifdef _CCI_UNIX_
  sem_WLock();
  #else
  WaitForSingleObject(hSemaphore,INFINITE);
  #endif

  memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
  /*Log("    ####control %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);*/
  ret=fwrite(hDxMemView,1,c.DxPoolUsedSize,fp);
  if(ret!=c.DxPoolUsedSize)
  {
          /*take action.. if not able to read full file*/
  }
  Log("    return of fwrite %d\n",ret);
  fclose(fp);
  #ifdef _CCI_UNIX_
  sem_WUnLock();
  #else
  ReleaseSemaphore(hSemaphore,1,NULL);
  #endif

  cciLog(&ret, CCI_LOG_ERROR, __FILE__, __LINE__, "storeDxPoolToFile", constDxCatalog, DXP_MEMORY_SAVED_TO_FILE, "MQSeriesIntegrator2", (char *)0);
  return DXP_SUCCESS;
}

/******************************************************************************/
/* initDxPoolFromFile                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  None                                                                      */
/*                                                                            */
/* Returns:                                                                   */
/*  0 - success                                                               */
/*  1 - failure                                                               */
/*                                                                            */
/* Description:                                                               */
/*  fills the shared Memory from file c:\DxPoolData.bin                       */
/******************************************************************************/
int initDxPoolFromFile()
{
  FILE *fp=NULL;
  int lSize;
  int ret;
  DX_POOL_CONTROL c;
  int rc;
  Log("-->initDxPoolFromFile\n");
#ifdef _CCI_UNIX_
  fp = fopen("/tmp/DxPoolData.bin","rb");
#else
  fp = fopen("c:\\DxPoolData.bin","rb");
#endif
  if(fp==NULL)
  {
        Log("  > File DxPoolData.bin does not exist or could not be opened");
        return DXP_FAILED;
  }
  fseek (fp , 0 , SEEK_END);
  lSize = ftell (fp);
  rewind (fp);
  if(lSize <= sizeof(DX_POOL_CONTROL)) /*if file size is less then control size*/
  {
        fclose(fp);
    return DXP_FAILED;
  }
  ret=fread(&c,1,sizeof(DX_POOL_CONTROL),fp);
  rewind (fp);
  if(ret!=sizeof(DX_POOL_CONTROL)) /* if not able to read the req bytes*/
  {
        fclose(fp);
        return DXP_FAILED;
  }
  if(c.DxPoolUsedSize!=lSize) /*if file size if not equal to the used size*/
  {
        fclose(fp);
        return DXP_FAILED;
  }

  #ifdef _CCI_UNIX_
  sem_WLock();
  #else
  WaitForSingleObject(hSemaphore,INFINITE);
  #endif

  memcpy(&c,hDxMemView,sizeof(DX_POOL_CONTROL));
  /*Log("    #control %d %d\n",c.DxPoolDataBlocks,c.DxPoolUsedSize);*/
  if(c.DxPoolUsedSize==sizeof(DX_POOL_CONTROL)) /*if the pool is not used??*/
  {
        if(c.DxPoolDataBlocks<DX_POOL_MEMORY_SIZE*noOfPoolPages) /* only if the file size is less then the pool*/
        {
                ret=fread(hDxMemView,1,lSize,fp);
                Log("    return of fwrite %d\n",ret);
                if(ret==lSize)
              cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "initDxPoolFromFile", constDxCatalog, DXP_MEMORY_INIT_FROM_FILE, "MQSeriesIntegrator2", (char *)0);
        }
  }
  fclose(fp);
  #ifdef _CCI_UNIX_
  sem_WUnLock();
  #else
  ReleaseSemaphore(hSemaphore,1,NULL);
  #endif
  Log("  Shared Memory updated with the file contents\n");

  return DXP_SUCCESS;
}
