/******************************************************************************/
/*                                                                            */
/* Progarm name    : DxPool.h                                                 */
/*                                                                            */
/* Description     : This header file contains function prototypes and        */
/*                   constants for DxPool.c                                   */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 2.0                                                                */
/* Author: Prafull Kumar                    Date: March 2003                  */
/*                                                                            */
/******************************************************************************/

/* constants */

#ifdef __cplusplus
extern "C" {
#endif

#include <BipCni.h>
#include "DxPoolUtil.h"
#include "node_utils.h"

#ifndef max
#define max(x, y) (((x) > (y)) ? (x) : (y))
#endif

#ifndef min
#define min(x,y) (((x)<(y)) ? (x) : (y))
#endif

/* ---------------------------------------------------------------- */
/*                     Internal constants                           */
/*                                                                  */
/* The interface expects some CciChar strings to be passed. These   */
/* have to be constructed from char strings using the supplied      */
/* utilities. The char constants are defined here along with the    */
/* static CciChar variables which will contain these constants      */
/* converted into CciChar format.                                   */
/*                                                                  */
/* ---------------------------------------------------------------- */

#define CONST_DXPOOL_NODE                    "DxPoolNode"
const CciChar * constDxPoolNode             = 0;

#define CONST_PLUGIN_NODE_FACTORY            "DxPoolNodeFactory"
const CciChar * constPluginNodeFactory      = 0;

#define CONST_OUT                            "out"
const CciChar * constOut                    = 0;

#define CONST_FAILURE                        "failure"
const CciChar * constFailure                = 0;

#define CONST_DXCATALOG                      "MQSI-DxPool"
CciChar * constDxCatalog                    = 0;

#define CONST_MQSI2                          "MQSeriesIntegrator2"
CciChar * constMQSI2                        = 0;

char    * constLabel                        = 0;

#define MAX_ATTR_NAME_LEN     128
#define MAX_NODE_NAME_LEN     128
#define MAX_STRING_SIZE       256
#define MAX_KEY_SIZE	      256
#define MAX_TERMINAL_NAME_LEN 256
#define CNI_TYPE_INTEGER        1
#define CNI_TYPE_CHAR           2
#define CNI_TYPE_BOOL           3
#define DX_POOL_MEMORY_NAME		"DxPoolNodeSharedMemory"
#define DX_POOL_SEMAPHORE_NAME          "DxPoolNodeSemaphore"

#define DX_POOL_MEMORY_SIZE		4096
#define DXP_FAILED				1
#define DXP_SUCCESS				0

/*structure to store the data*/
typedef struct DxPoolDataBlock
{
  int   totalLength;
  int   keyLength;
  int   dataLength;
  char* dataKey;
  char* data;
} DX_POOL_DATA;

/*structure to control the data structure*/
typedef struct DxPoolControl
{
  int poolSize;
  int DxPoolDataBlocks;
  int DxPoolUsedSize;
} DX_POOL_CONTROL;

/*will store each line of the DxPoolExpression*/
typedef struct DxPoolExpressions
{
  void    *next;                         /* Pointer to next entry */
  void    *valuePtr;				     /* Pointer to value */
} DX_POOL_EXPRESSIONS;

#define CCI_CHECK_RC()                                                         \
   if (rc != CCI_SUCCESS)                                                      \
   {                                                                           \
     if (rc == CCI_EXCEPTION) cniRethrowLastException(0);                      \
     else                                                                      \
         {                                                                         \
       cniLog(                                                                 \
              &rc                                                              \
             ,CCI_LOG_ERROR                                                    \
                         ,__FILE__                                                         \
                         ,__LINE__                                                         \
                         ,""                                                               \
                         ,0                                                                \
                         ,"Message not propagated"                                         \
                         ,0                                                                \
                         );                                                                \
       return;                                                                 \
     }                                                                         \
   }

typedef struct attributeTblEntry {
  void*    next;                    /* Pointer to next entry */
  void*    valptr;                  /* Pointer to value */
  int      type;                    /* Value type */
  CciChar  name[MAX_ATTR_NAME_LEN]; /* Attribute name */
} ATTRIBUTE_TBL_ENTRY;

typedef struct terminalListEntry {
  void*        next;                        /* Pointer to next entry */
  CciTerminal* handle;                      /* Handle to terminal object */
  CciChar      name[MAX_TERMINAL_NAME_LEN]; /* Terminal name */
} TERMINAL_LIST_ENTRY;

typedef struct context {
  CciNode*             nodeObject;                  /* Handle for node object */
  ATTRIBUTE_TBL_ENTRY* attrListHead;                /* Head of linked list of attributes */
  ATTRIBUTE_TBL_ENTRY* attrListPrevious;            /* Previous entry in attribute list */
  TERMINAL_LIST_ENTRY* inputTerminalListHead;       /* Head of linked list of input terminal handles */
  TERMINAL_LIST_ENTRY* inputTerminalListPrevious;   /* Previous entry in input terminal list */
  TERMINAL_LIST_ENTRY* outputTerminalListHead;      /* Head of linked list of output terminal handles */
  TERMINAL_LIST_ENTRY* outputTerminalListPrevious;  /* Previous entry in output terminal list */
  CciChar              nodeName[MAX_NODE_NAME_LEN]; /* Node name */
  DX_POOL_EXPRESSIONS*  pDxPoolExpressions;
} CONTEXT_STRUCTURE;

/* Function prototypes */

ATTRIBUTE_TBL_ENTRY *insAttrTblEntry(
  CONTEXT_STRUCTURE* context,
  CciChar*         attrName,
  int              type
);

void *getInputTerminalHandle(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
);

TERMINAL_LIST_ENTRY *insInputTerminalListEntry(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
);

void *getOutputTerminalHandle(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
);

TERMINAL_LIST_ENTRY *insOutputTerminalListEntry(
  CONTEXT_STRUCTURE* context,
  CciChar*         terminalName
);

void traverseMessage
(
  void *element,
  FILE* outputFile
);

void test(void *element);

CciContext* createNodeContext(
  CciFactory* factoryObject,
  CciChar*    nodeName,
  CciNode*    nodeObject
);

int getAttributeName(
  CciContext* context,
  int         index,
  CciChar*    buffer,
  int         bufsize
);

int setAttribute(
  CciContext* context,
  CciChar*    attrName,
  CciChar*    attrValue
);

int getAttribute(
  CciContext* context,
  CciChar*    attrName,
  CciChar*    buffer,
  int         bufsize
);

void evaluate(
  CciContext* context,
  CciContext* destinationList,
  CciContext* exceptionList,
  CciContext* message
);

void* getAttributePointer(void *context,const CciChar *attrName);

void deleteNodeContext(void* context);

int initHandleOfDxPool();
void closeHandleOfDxPool();
void initHandleOfDxPoolView(int);
void closeHandleOfDxPoolView();
void Log(char *szFormat, ...);
void ucsLog(char* fmt,CciChar *ucsStr);

void initDxPoolMemory();
int storeDxPoolDataBlock(DX_POOL_DATA*);
int getDxPoolDataBlock(DX_POOL_DATA*,char*);
int processDxPoolscript(CONTEXT_STRUCTURE* context);
int  runDxPoolscript(CONTEXT_STRUCTURE* context,CciMessage*,
					 CciMessage*,CciMessage*,CciMessage*,char**);
int insertDxScript(CONTEXT_STRUCTURE*, char*);

int getOperandsOfPoolScript(char*,char*,char*);
int getBoolAttribute(void *context,const CciChar *attrName);

int storeDxPoolToFile();
int initDxPoolFromFile();
#ifndef _CCI_UNIX_
#endif

#ifdef _CCI_UNIX_
void sem_WLock();
void sem_RLock();
void sem_WUnLock();
void sem_RUnLock();
#endif

#ifdef __cplusplus
}
#endif

