/******************************************************************************/
/*                                                                            */
/* Progarm name    : DxPool.h                                                 */
/*                                                                            */
/* Description     : This header file contains function prototypes and        */
/*                   constants for DxPool.c                                   */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Prafull Kumar                    Date: June 2001                   */
/*                                                                            */
/******************************************************************************/

/* constants */

#define MAX_ATTRIBUTE_NAME_LEN     128
#define MAX_NODE_NAME_LEN          128
#define MAX_TERMINAL_NAME_LEN      256
#define MAX_STRING_SIZE			   256
#define MAX_KEY_SIZE			   256
#define MAX_DX_SCRIPT_LINE_SIZE    256

#define DX_POOL_MEMORY_NAME		"DxPoolNodeSharedMemory"
#define DX_POOL_SEMAPHORE_NAME  "DxPoolNodeSemaphore"
#define DX_POOL_MEMORY_SIZE		4096
#define DXP_FAILED				1
#define DXP_SUCCESS				0


#define BIP2623 ((DWORD)0x40000A3FL)

#ifndef min                                   
#define min(x,y) (((x)<(y)) ? (x) : (y))      
#endif                                        

#define CCI_CHECK_RC()                                                         \
   if (rc != CCI_SUCCESS)                                                      \
   {                                                                           \
     if (rc == CCI_EXCEPTION) cniRethrowLastException(0);                      \
     else                                                                      \
	 {                                                                         \
       cniLog(                                                                 \
              &rc                                                              \
             ,CCI_LOG_ERROR                                                    \
			 ,__FILE__                                                         \
			 ,__LINE__                                                         \
			 ,""                                                               \
			 ,0                                                                \
			 ,"Message not propagated"                                         \
			 ,0                                                                \
			 );                                                                \
       return;                                                                 \
     }                                                                         \
   }
/*structure to store the data*/
typedef struct DxPoolDataBlock
{
  int   totalLength;
  int   keyLength;
  int   dataLength;
  char* dataKey;
  char* data;
} DX_POOL_DATA;

/*structure to control the data structure*/
typedef struct DxPoolControl
{
  int poolSize;
  int DxPoolDataBlocks;
  int DxPoolUsedSize;
} DX_POOL_CONTROL;

/*will store each line of the DxPoolExpression*/
typedef struct DxPoolExpressions
{
  void    *next;                         /* Pointer to next entry */
  void    *valuePtr;				     /* Pointer to value */
} DX_POOL_EXPRESSIONS;

/* typedefs */

typedef struct attributeTableEntry
{
  void    *next;                         /* Pointer to next entry */
  void    *valuePtr;                     /* Pointer to value */
  int      type;                         /* Value type */
  wchar_t  name[MAX_ATTRIBUTE_NAME_LEN]; /* Attribute name */
} ATTRIBUTE_TABLE_ENTRY;

typedef struct terminalListEntry
{
  void    *next;                        /* Pointer to next entry */
  void    *handle;                      /* Handle to terminal object */
  wchar_t  name[MAX_TERMINAL_NAME_LEN]; /* Terminal name */
} TERMINAL_LIST_ENTRY;

typedef struct context
{
  void *nodeObject;                                 /* Handle for node object */
  ATTRIBUTE_TABLE_ENTRY* attributeListHead;         /* Head of linked list of attributes */
  ATTRIBUTE_TABLE_ENTRY* attributeListPrevious;     /* Previous entry in attribute list */
  TERMINAL_LIST_ENTRY* inputTerminalListHead;       /* Head of linked list of input terminal handles */
  TERMINAL_LIST_ENTRY* inputTerminalListPrevious;   /* Previous entry in input terminal list */
  TERMINAL_LIST_ENTRY* outputTerminalListHead;      /* Head of linked list of output terminal handles */
  TERMINAL_LIST_ENTRY* outputTerminalListPrevious;  /* Previous entry in output terminal list */
  wchar_t nodeName[MAX_NODE_NAME_LEN];              /* Node name */
  DX_POOL_EXPRESSIONS*  pDxPoolExpressions;
} CONTEXT_STRUCTURE;


/* Function Prototypes */

/* Node Utility Functions */
ATTRIBUTE_TABLE_ENTRY *insertAttributeTableEntry
(
  CONTEXT_STRUCTURE *context,
  wchar_t    *nodeName, 
  wchar_t    *attributeName,
  int         type
);

void *getInputTerminalHandle
(
  CONTEXT_STRUCTURE *context,
  wchar_t    *terminalName
);

TERMINAL_LIST_ENTRY *insertInputTerminalListEntry
(
  CONTEXT_STRUCTURE *context,
  wchar_t *terminalName 
);

void *getOutputTerminalHandle
(
  CONTEXT_STRUCTURE *context,
  wchar_t *terminalName
);

TERMINAL_LIST_ENTRY *insertOutputTerminalListEntry
(
  CONTEXT_STRUCTURE *context,
  wchar_t *terminalName 
);

void traverseMessage
(
  void *element,
  FILE* outputFile
);

void test(void *element);

/* Node Implementation Functions */
void *createNodeContext(
                        void    *factoryObject,
                        wchar_t *nodeName,
                        void    *nodeObject
                       );

int getAttributeName(
                     void*    context,
                     int      index,
                     wchar_t* buffer,
                     int      bufferSize
                    );

int setAttribute(
                 void    *context,
                 wchar_t *attributeName,
                 wchar_t *attributeValue
                );

int getAttribute(
                 void    *context,
                 wchar_t *attributeName,
                 wchar_t *buffer,
                 int      bufferSize
                );

void evaluate(
              void *context, 
              void *destinationList,
              void *exceptionList,
              void *message
             );
void* getAttributePointer(
                 void *context
				,wchar_t *attrName
                );

void deleteNodeContext(void* context);

int initHandleOfDxPool();
void closeHandleOfDxPool();
void initHandleOfDxPoolView(int);
void closeHandleOfDxPoolView();
void log(char *szFormat, ...);

void initDxPoolMemory();
int storeDxPoolDataBlock(DX_POOL_DATA*);
int getDxPoolDataBlock(DX_POOL_DATA*,char*);
int processDxPoolscript(CONTEXT_STRUCTURE* context);
int  runDxPoolscript(CONTEXT_STRUCTURE* context,CciMessage*,
					 CciMessage*,CciMessage*,CciMessage*,char**);
int insertDxScript(CONTEXT_STRUCTURE*, char*);

int getOperandsOfPoolScript(char*,char*,char*);
int getBoolAttribute(void *context,wchar_t *attrName);

int storeDxPoolToFile();
int initDxPoolFromFile();
